#!/usr/bin/env node

const unirest = require('unirest')
const fs = require('fs')
const path = require('path')
const url = require('url')
const colors = require('colors')
const chokidar = require('chokidar')
const yargs = require('yargs')
const headWatcher = require('./lib/head_watcher')
const utils = require('./lib/utils')
const glob = require('glob')
const co = require('co')
const pkg = require('./package.json')
const throttle = utils.throttle
const logger = utils.logger

let zedUrl

try {
  utils.notifyUpdate()
} catch (e) {
  logger.error(e)
}

process.on('uncaughtException', logger.error)

function handleResponse(relativePath) {
  return function (response) {
    if (!response || !response.body) {
      return logger.log(colors.red('Oops, no response...'), relativePath)
    }
    if (response.error) {
      logger.error(response.body, relativePath)
    } else {
      logger.log(colors.green(response.body), relativePath)
    }
  }
}

function uploadFile(localFile) {
  const relativePath = path.relative(process.cwd(), localFile)
  const data = fs.readFileSync(localFile, 'utf-8')
  // eslint-disable-next-line no-use-before-define
  const putUrl = url.resolve(zedUrl + '/', relativePath)
  logger.log('PUT'.green, putUrl)
  unirest.put(putUrl)
    .headers({
      'Content-Type': 'application/octet-stream'
    })
    .send(data)
    .end(handleResponse(relativePath))
}

function deleteFile (localFile) {
  const relativePath = path.relative(process.cwd(), localFile)
  // eslint-disable-next-line no-use-before-define
  const deleteUrl = url.resolve(zedUrl + '/', relativePath)
  logger.log('DELETE'.red, deleteUrl)
  unirest.delete(deleteUrl)
    .headers({
      'Content-Type': 'application/octet-stream'
    })
    .end(handleResponse(relativePath))
}

function handleChange (method, localFile) {
  const relativePath = path.relative(process.cwd(), localFile)

  if (headWatcher.isChecking()) {
    logger.log('Skip on checking out '.yellow + headWatcher.getCurrentBranch().yellow, relativePath)
    return
  }
  if (!headWatcher.isOriginalBranch()) {
    logger.log('Skip on branch '.yellow + headWatcher.getCurrentBranch().yellow, relativePath)
    return
  }

  let handleFile = uploadFile
  if (method === 'delete') {
    handleFile = deleteFile
  }
  throttle(handleFile.bind(null, localFile))
}

const defaultIgnoredPaths = [/[\/\\]\./, 'venv/*', /node_modules|bower_components/]
function coerceIgnore(files) {
  // console.log('args', files)
  // looks like there has bugs
  if (Array.isArray(files[files.length - 1])) {
    files.pop()
  }
  const patterns = files.map(function (file) {
    if (file instanceof RegExp) return file
    const match = file.match(/^\/(.*)\/($|i|m|g|u|y)/)
    if (match) {
      file = new RegExp(match[1], match[2])
    }
    return file
  })
  return patterns
}


/* check comand options start */
function verify(argv) {
  if (!argv.url && !argv.pre) {
    logger.errorAndExit('Please specify zed server url or pre to automatically get one!\n', 'e.g. \n`zs -u http://hador5.intra.douban.com:7337/fs/1ecf197a687a40fd9affac65ffaba514`\nor\n`zs -p yangzhen@155`')
  }

  if (!fs.existsSync(process.cwd() + '/.git')) {
    logger.errorAndExit('Please run this command in your project root!\n',
    'Current dir is:\n' + process.cwd())
  }

  if (argv.branch) {
    const curBranch = headWatcher.getBranchFromFile()
    if (curBranch !== argv.branch) {
      logger.errorAndExit('Branch doesn\'t match ', '(-b ' + argv.branch + ' !== ' + curBranch + ')')
    }
  }

  if (argv.up && !argv.pre) {
    logger.errorAndExit('If you want to update your pre before update, `-p/--pre` must be specified.')
  }
}

/* check comand options end */

function* initSync(argv) {
  zedUrl = argv.url
  if (!argv.url && argv.pre) {
    if (argv.up) {
      const [user, pre] = argv.pre.split('@')
      yield utils.SSHExec(user, `dae pre update --pre ${pre}`)
      yield utils.sleep(100)
      utils.cache.remove(`url:${argv.pre}`)
    }
    zedUrl = argv.url = yield utils.getZedURLByPreInfo(argv.pre, {force: argv.force})
  }

  if (argv.verify) {
    utils.validateZedURL(argv.url, headWatcher.getBranchFromFile())
  }
  if (argv.sync.length) {
    argv.sync.forEach(function (patt) {
      glob(patt, {
        follow: true,
        nodir: true
      }, function (err, files) {
        if (err) {
          logger.errorAndExit(err)
        }
        files.forEach(function (file) { handleChange('change', file) })
      })
    })
  }

  headWatcher.init(argv)

  logger.log('Start watching dirs: '.green + argv.watch)

  const codeWatcher = chokidar.watch(argv.watch, {
      ignored: argv.ignore,
      persistent: true,
      followSymlinks: !!argv.follow,
      ignoreInitial: true,
    })
    .on('add', handleChange.bind(this, 'add'))
    .on('change', handleChange.bind(this, 'change'))
    .on('unlink', handleChange.bind(this, 'delete'))
}

function clearCache(argv) {
  const oldKeys = utils.cache.keys()
  if (argv.clear && argv.clear.length) {
    argv.clear.forEach(key => utils.cache.removeKey(key))
  }
  if (argv.clearAll) {
    utils.cache.keys().forEach(key => utils.cache.removeKey(key))
  }
  if (argv.clearPrefixes.length) {
    utils.cache.keys()
      .filter(key => argv.prefixes.some(pre => key.startsWith(pre)))
      .forEach(key => utils.cache.removeKey(key))
  }
  const newKeys = utils.cache.keys()
  if (argv.cacheList) {
    logger.log('Existed keys:'.yellow , '\n' + newKeys.map(key => {
      const { ttl, expire } = utils.cache.getRaw(key)
      return `key: ${key}\nttl: ${ttl / 24 / 1000 / 3600 } days,    expire: ${new Date(expire).toLocaleString()}`
    }).join('\n'))
  } else {
    logger.log('Cleared keys:'.yellow, oldKeys.filter(key => !newKeys.includes(key)).join(','))
    logger.log('Existed keys:'.yellow, newKeys.join(','))
  }
}


const program = yargs
  .version(pkg.version)
  .usage(pkg.description + `
Usage:
  Watch Files:
    $0 -p <yourDev2LoginName>@<preNumber>
    $0 -u <zedUrl>
  Clear cache:
    $0 --clear branch:<zedUrl>                  clear branch validation cache
    $0 --clear url:<yourName>@<yourPreNum>      clear pre url cache
    $0 --cache-list                             list cache
`)
  .option('u', {
    alias: 'url',
    // demandOption: true,
    // default: false,
    describe: 'Zed url, plz see http://readthedocs.dapps.douban.com/docs/dae/dev/pre.html#id2',
    type: 'string',
  })
  .option('p', {
    alias: 'pre',
    describe: 'The user and pre to get zed url, format: `<user>@<pre num>`, e.g. yangzhen@155',
    type: 'string',
  })
  .option('up', {
    alias: 'update-pre',
    default: false,
    describe: 'Update your pre before get zed url from the pre, -p/--pre required.',
    type: 'boolean',
  })
  .implies('up', 'p')
  .option('w', {
    alias: 'watch',
    default: [process.cwd()],
    describe: 'Directories to watching, default is current directory.',
    type: 'array',
  })
  .option('follow', {
    default: false,
    describe: 'whether follow symlinks when watching, default is false.',
    type: 'boolean',
  })
  .option('s', {
    alias: 'sync',
    default: [],
    describe: 'Upload files using glob pattern.',
    type: 'array',
  })
  .option('b', {
    alias: 'branch',
    describe: 'Just for command history and check.',
    type: 'string',
  })
  .option('v', {
    alias: 'verify',
    describe: 'Whether verify zed url, default is true.',
    type: 'boolean',
    default: true,
  })
  .option('f', {
    alias: 'force',
    describe: 'Force get new zed url from pre, this will ignore zed url cache.',
    type: 'boolean',
    default: false,
  })
  .implies('f', 'p')
  .option('i', {
    alias: 'ignore',
    default: defaultIgnoredPaths,
    coerce: coerceIgnore,
    describe: 'ignored paths when watching, support glob pattern and regex, default is "-i /[\/\\\\]\./ venv/* /node_modules|bower_components/ "',
    type: 'array',
  })
  .option('ld', {
    alias: 'local-dae',
    default: false,
    type: 'boolean',
    descript: 'Using dae command in local machine instead of dev2.',
  })
  .option('remote', {
    alias: 'r',
    default: 'doubandev3.intra.douban.com',
    type: 'remote',
    descript: 'remote server to execute dae command',
  })
  .option('clear', {
    default: [],
    describe: 'The keys to clear.',
    type: 'array',
  })
  .option('clear-all', {
    default: false,
    describe: 'Clear all keys.',
    type: 'boolean',
  })
  .option('clear-prefixes', {
    default: [],
    describe: 'Clear keys by prefixes.',
    type: 'array',
  })
  .option('cache-list', {
    default: false,
    describe: 'List cache keys',
    type: 'boolean',
  })
  .showHelpOnFail(true, 'Specify --help for available options')
  .help('h')
const { argv } = program

logger.argv = argv

if (process.argv.length <= 2) {
  program.showHelp()
  process.exit(0)
}
if (argv.clear.length || argv.clearAll || argv.clearPrefixes.length || argv.cacheList) {
  clearCache(argv)
} else {
  if (argv.verify) {
    verify(argv)
  }
  co.wrap(initSync)(argv).catch(console.error)
}
